package io;

import java.io.FileInputStream;
import java.util.StringTokenizer;
import java.util.Vector;

import wwc.messaging.Theater;
import io.core.agents.AgentFactory;
import io.resources.InterfaceActor;

public class IOTheater extends Theater {
	public static String decisionAgent = null;
	public static String profilingAgent = null;
	public static String protocolAgent = null;
	private static Vector peerServers = new Vector();

	private static boolean userInterface = false;
	private static String port = null;

	public static boolean silent = false;

	public static void main(String[] arguments) {
		getSettings();
		AgentFactory.createFactory(decisionAgent, profilingAgent, protocolAgent);
		parseArguments(arguments);

		if (port == null) helpMessage();
		String[] newArgs = { port };
		Theater.main(newArgs);

		AgentFactory.getDecision().start();

		if (!peerServers.isEmpty()) AgentFactory.getProtocol().joinComputation(peerServers);
		if (userInterface) new UserInterface().start();

		InterfaceActor i = new InterfaceActor();
	}

	private static void parseArguments(String[] args) {
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-?") || args[i].equals("-h"))       helpMessage();
			else if (args[i].equals("-p") || args[i].equals("-u") || args[i].equals("-f") || args[i].equals("-silent")) {
				try {
					if (args[i].equals("-p"))               port = args[i+1];
					else if (args[i].equals("-f"))          getArgsFromFile(args[i+1]);
					else if (args[i].equals("-u"))		userInterface = true;
					else if (args[i].equals("-silent"))	silent = true;
				} catch (Exception e) {
					if (args[i].equals("-p"))               helpMessage("invalid port number : " + args[i+1]);
				}
				i = i + 1;
			} else if (args[i].equals("-s")) {
				i++;
				while ((i < args.length) && ((args[i]).charAt(0) != '-')) {
					peerServers.add(args[i]);
					i++;
				}
				i--;
				if (args[i].equals("-s")) helpMessage("You must specify at least one PeerServer");
			} else helpMessage("Invalid Arguments");
		}
	}

        private static void getArgsFromFile(String filename) {
                FileInputStream file;
                byte input[];
                                        
                try {
                        file = new FileInputStream(filename);
                                
                        input = new byte[file.available()];
                        file.read(input);
                } catch (Exception e) {
                        System.err.println("File Error: " + e);
                        return;
                }
                         
                StringTokenizer st = new StringTokenizer(new String(input), "\n ");
                String args[] = new String[st.countTokens()];
        
                int i = 0;
                while (st.hasMoreTokens()) {
                        args[i] = st.nextToken();   
                        i++;
                }
        }

	private static void getSettings() {
		String settingsFile = System.getProperty( "settings" );
		if (settingsFile == null) return;

		byte input[];

		try {
			FileInputStream file = new FileInputStream(settingsFile);
			input = new byte[file.available()];
			file.read(input);
                
			StringTokenizer st = new StringTokenizer(new String(input), "\n");
                        
			decisionAgent = st.nextToken();
			profilingAgent = st.nextToken();
			protocolAgent = st.nextToken();

			while (st.hasMoreTokens()) {
				peerServers.add(st.nextToken());
			}
		} catch (Exception e) {
			System.err.println("Error reading IO theater information file: " + e);
			System.err.println("\tSettings File Format: ");
			System.err.println("\t<decision agent class name><newline>");
			System.err.println("\t<profiling agent class name><newline>");
			System.err.println("\t<protocol agent class name><newline>");
			System.err.println("\t[<name server location><newline>]*");
			System.exit(0);
		}
	}

        private static void helpMessage(String s) {
                System.out.println(s);  
                helpMessage();
        }                       
                        
        private static void helpMessage() {
                System.out.println("usage :");
                System.out.println("\tjava ...PeerTheater"); 
                System.out.println("\tjava ...PeerTheater -h");
                System.out.println("\tjava ...PeerTheater -p   portNumber");
                System.out.println("\t                    -u");
                System.out.println("\t                    -f   filename");
                System.out.println("\t                    -s   server1 server2 ... serverN");
	
                System.out.println("options :");
                System.out.println("\t-h                             : print this message");
                System.out.println("\t-u                             : run with a user interface");
                System.out.println("\t-f filename                    : load settings from file");
                System.out.println("\t                                 file is in the same format as command line arguments");
                System.out.println("\t-p portNumber                  : set the listening port to portNumber");
                System.out.println("\t                                 default port number is 3030");
                System.out.println("\t-s server1 server2 ... serverN : gets connections to other");
                System.out.println("\t                                 PeerTheaters, from servers 1 through N");

		System.exit(1);
        }

}
