package io.ars.decision;

import io.IOTheater;
import io.ars.profiling.ARSProfilingAgent;
import io.ars.profiling.ActorProfile;
import io.ars.protocol.ARSProtocolAgent;
import io.ars.protocol.ARSteal;
import io.core.agents.AgentFactory;
import io.core.agents.DecisionAgent;

import java.util.Vector;

import salsa.language.ServiceFactory;
import salsa.messaging.TheaterService;
import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;


public class ARSDecisionAgent extends Thread implements DecisionAgent {

	public boolean computing = true;
	private boolean waiting = false;
	private int sleepTime = 5000;
	private long lastReset = System.currentTimeMillis();

	private ARSProfilingAgent profilingAgent = (ARSProfilingAgent)AgentFactory.getProfiling();
	private ARSProtocolAgent protocolAgent = (ARSProtocolAgent)AgentFactory.getProtocol();
	private double processing = profilingAgent.getProcessing();

	TheaterService theaterService = ServiceFactory.getTheater();
	NamingService namingService = ServiceFactory.getNaming();

	public void stealTerminated()		{ waiting = false; }
	public boolean computing()		{ return computing; }
	public double usedProcessing()		{ return processing/profilingAgent.getActorProfiles().size(); }
	public double availableProcessing()	{ return processing/(profilingAgent.getActorProfiles().size()+1); }

	public void leaveComputation() {
		computing = false;
	}
	public void enterComputation() {
		computing = true;
	}

	public double decisionFunction(ActorProfile candidate, ARSteal steal) {
		String localId = protocolAgent.getLocation().toString() + "/";
		String foreignId = steal.getOrigin().toString() + "/";

		double gain;
		gain = candidate.processed() * (steal.getProcessing()/usedProcessing());
		gain += (candidate.sentTo(foreignId) - candidate.sentTo(localId));
		gain += (candidate.receivedFrom(foreignId) - candidate.receivedFrom(localId));
		gain /= candidate.processed();
		gain -= 1;
		return gain;
	}

	public ActorProfile getBestCandidate(ARSteal stealPacket) {
		ActorProfile candidate = null;
		double candidateGain = 0;

		Vector actorProfiles = profilingAgent.getActorProfiles();

		for (int i = 0; i < actorProfiles.size(); i++) {
			ActorProfile current = (ActorProfile)actorProfiles.get(i);

			if (!current.scheduled()) {
				double gain = decisionFunction(current, stealPacket);

				if (gain > candidateGain) {
					candidate = current;
					candidateGain = gain;
				}
			}
		}

		return candidate;
	}

	private void beginSteal() {
		ARSteal steal = new ARSteal(protocolAgent.getLocation(), availableProcessing());
		protocolAgent.beginSteal(steal);
	}

	public void resolveActors() {
                Vector keys = new Vector(theaterService.getRegistry().keySet());

                for (int i = 0; i < keys.size(); i++) {
                        Object current = keys.get(i);

                        if (theaterService.getRegistry().get( current ) instanceof UAL) {
				try {
					theaterService.getRegistry().put(current, namingService.resolve( new UAN((String)current) ));
				} catch (Exception e) {
					System.err.println("Exception inside ARSDecisionAgent.resolveActors, should not happen");
				}
                        }
                }
	}


	public void run() {
		if (!IOTheater.silent) System.out.println("Time\t\tProcessed\tActors");
		try {
			sleep(sleepTime);
		} catch (InterruptedException ie) {
			System.err.println("Sleep interrupted!\n");
		}

		while (true) {          
			if (profilingAgent.isLight() && !waiting && computing) beginSteal();

			if (!computing) protocolAgent.leaveComputation();
			else resolveActors();

			if (!IOTheater.silent) System.out.println(System.currentTimeMillis() + "\t" + profilingAgent.processed() + "\t\t" + profilingAgent.getActorProfiles().size()); 
			profilingAgent.reset();
                
			try {
				sleep(sleepTime);
			} catch (InterruptedException ie) {
				System.err.println("Sleep interrupted!\n");
			}
		}
	}               
}

