package salsac;
/**
 * $Id: SalsaCodeGenerator.java,v 1.25 2003/06/24 01:12:27 deselt Exp $
 *
 * This is the Salsa programming language code generator.
 * We use the abstract syntax tree produced by the parsing phase to build
 * Java files corresponding to the original Salsa files being compiled.
 *
 * Author: Carlos Varela	cvarela@uiuc.edu	1999/09/14.
 *
 */
import java.io.*;

public class SalsaCodeGenerator {
	 SalsaNode tree;
	 PrintWriter writer;
	 static String behaviorName;
	 static String extendedBehaviorName;
	 static boolean generateMain = false;
	 static String indent;

	SalsaCodeGenerator(SalsaNode node, OutputStream stream) {
		this.tree = node;
		this.writer = new PrintWriter(stream);
		indent = "\t\t";
	}

	public static void setIndent(String _indent) {
		indent = _indent;
	}


	public void generateCode() throws CodeGenerationException {
		tree.generateJavaCode(writer);
		writer.flush();
		writer.close();
	}

	public static String generateMessageSend(MessageSendingData msData) {

		return generateMessageSend(msData.getTarget(), msData.getTargets(),
					   msData.getMethodNames(), msData.getArguments(),
					   msData.getTokens(), msData.getSender(),
					   true, msData.getTokenClass(), msData.getFirstClass());
	}

	 public static String generateMessageSend(MessageSendingData msData, boolean withSend) {

		return generateMessageSend(msData.getTarget(), msData.getTargets(),
					   msData.getMethodNames(), msData.getArguments(),
					   msData.getTokens(), msData.getSender(),
					   withSend, msData.getTokenClass(), msData.getFirstClass());
	 }

	 public static String generateMessageSend(String target, String targets, String methodNames,
						  String arguments, String tokens, String sender,
						  boolean withSend, String tokenClass, boolean firstClass) {
		String messageSendString = "";

		if (withSend) {
			messageSendString = 	"{\n" +
						indent + "\tActorReference[] _targets = { " + targets + " };\n" +
						indent + "\tString[] _methodNames = { " + methodNames + " };\n" +
						indent + "\tObject[][] _arguments = { " + arguments + " };\n\n";

			if (firstClass) {
				messageSendString += indent + "\tMessage _continuation = new Message( " + sender + ", _targets, _methodNames, _arguments );\n";
				messageSendString += indent + "\t_continuation.updateFirstClass(currentContinuation);\n";
				messageSendString += indent + "\t" + target + ".send(_continuation);\n";
			} else {			
				messageSendString += indent + "\t" + target + ".send( " + 
						     "new Message( " + sender + ", _targets, _methodNames, _arguments ) );\n";
			}

			messageSendString +=	indent + "}\n";
		}
		else {
			messageSendString =	indent + "\n" +
						indent + "\tActorReference[] _targets = { " + targets + " };\n" +
						indent + "\tString[] _methodNames = { " + methodNames + " };\n" +
						indent + "\tObject[][] _arguments = { " + arguments + " };\n\n" +
						indent + "\tMessage " + target + " = " +
							 "new Message( " + sender + ", _targets, _methodNames, _arguments );\n";

			if (firstClass) messageSendString += indent + "\t" + target + ".updateFirstClass(currentContinuation);\n";

			messageSendString +=	indent + "\n";
		}

		return messageSendString;
	 }

	 public static void setBehaviorName(String behavior) {
		behaviorName = behavior;
	 }

	 public static String generateMainMethod() {
	// cvarela: create main that reports no "act" found at run-time,
	//			 or give a compiler warning...
		if (!generateMain)
			 return "";

		StringBuffer mainMethodCode = new StringBuffer();
		String behaviorInstance = behaviorName.substring(0,1).toLowerCase()
			 +behaviorName.substring(1,behaviorName.length());
		mainMethodCode.append("\n\tpublic static void main(String args[]) {\n");
		mainMethodCode.append("\t\t"+behaviorName+" _"+behaviorInstance
						+" = new "+ behaviorName+"();\n\n\t\t");
		// _behaviorName<-act(args);

		indent += "\t";
		MessageSendingData actSend = new MessageSendingData("null",
								    "_"+behaviorInstance,
								    "_"+behaviorInstance,
								    "\"act\"",
								    "{ args }",
								    "-1");


		mainMethodCode.append(generateMessageSend(actSend));

		mainMethodCode.append("\n\t}\n");
		return new String(mainMethodCode);
	 }

	 public static String generateJoinStatement(JoinStatementData jsData) throws CodeGenerationException {
		return generateJoinStatement(jsData.getSender(), 
					     jsData.getTarget(),
					     jsData.getTargets(),
					     jsData.getMethodNames(),
					     jsData.getTokens(),
					     jsData.getArguments(),
					     jsData.getContinuationExpression(),
					     jsData.getNumber(),
					     jsData.getWithSend(),
					     jsData.getActorArray(),
					     jsData.getWithContinuation());
	 }

	public static String generateJoinStatement(String sender,
						    String target,
						    String targets,
						    String methodName,
						    String tokenPosition,
						    String arguments,
						    SalsaNode continuationExpression,
						    int number,
						    boolean withSend,
						    boolean actorArray,
						    boolean withContinuation) throws CodeGenerationException {
		StringBuffer joinCode = new StringBuffer();
		joinCode.append(generateJoinStatementPre(target, targets, methodName, tokenPosition, arguments, number, actorArray));
		joinCode.append(generateJoinStatementDirector(sender, target, number));
		joinCode.append(generateJoinStatementContinuationCode(continuationExpression, number+1, target));
		joinCode.append(indent + "\t" + target + ".setContinuation(_continuation" + number + ");\n");
		joinCode.append(indent + "}\n");
		if (withContinuation) {
			joinCode.append(indent + "//  "+target+"<-process();\n" + indent);

			MessageSendingData processSend = new MessageSendingData( "self",
										 (withSend?target:"_continuation"+(number-1)),
										 target,
										 "\"process\"",
										 "{ }",
										 "-1");

			joinCode.append(generateMessageSend(processSend, withSend) + "\n");
		}

		return joinCode.toString();
	}

	public static String generateJoinStatementPre(String target,
						      String targets,
						      String methodNames,
						      String tokenPositions,
						      String arguments,
						      int number,
						      boolean actorArray) {
		StringBuffer joinCode = new StringBuffer();
		joinCode.append("JoinDirector "+target+" = null;\n");
		joinCode.append(indent + "{\n");
		if (actorArray) {
			 joinCode.append(indent + "\tActorReference[] _targets"+number+" = "+targets+";\n");
			 joinCode.append(indent + "\tString[] _methodNames"+number+" = new String[" + targets + ".length];\n");
			 joinCode.append(indent + "\tint[] _tokenPositions" + number + " = new int[" + targets + ".length];\n");
			 joinCode.append(indent + "\tObject [][] _arguments"+number+" = new Object[" + targets + ".length][];\n");
			 joinCode.append(indent + "\tfor (int i = 0; i< " + targets + ".length; i++){\n");
			 joinCode.append(indent + "\t\t_methodNames"+number+"[i] = "+methodNames+";\n");
			 joinCode.append(indent + "\t\t_tokenPositions"+number+"[i] = " + tokenPositions + ";\n");

			 if (arguments.equals( "{}" )) {
				joinCode.append(indent + "\t\t_arguments"+number+"[i] = ");
			 	joinCode.append("new Object[0];\n");
			 } else {
				joinCode.append(indent + "\t\tObject[] temp = " + arguments + ";\n");
				joinCode.append(indent + "\t\t_arguments"+number+"[i] = temp;\n");
			 }

			 joinCode.append(indent + "\t}\n");
		} else {
			 joinCode.append(indent + "\tActorReference[] _targets"+number+" = { "+targets+" };\n");
			 joinCode.append(indent + "\tString[] _methodNames"+number+" = { "+methodNames+" };\n");
			 joinCode.append(indent + "\tint[] _tokenPositions"+number+" = { "+tokenPositions+" };\n");
			 joinCode.append(indent + "\tObject [][] _arguments"+number+" = { "+arguments+" };\n");
		}
		return joinCode.toString();
	}

	public static String generateJoinStatementContinuationCode(SalsaNode continuationExpression,
								   int number,
								   String target) throws CodeGenerationException {
	
		return continuationExpression.generateContinuationExpressionCode(number,
										 false,
										 "_continuation"+(number-1),
										 target.getClass().getName()/*+".getReturnType()"*/);
	}

	public static String generateJoinStatementDirector(String sender, String director, int number) {
		StringBuffer joinCode = new StringBuffer();
		joinCode.append(indent + "\t"+director+" = new JoinDirector(");
		joinCode.append(sender+".self, ");
		joinCode.append("_targets"+number+", ");
		joinCode.append("_methodNames"+number+", ");
		joinCode.append("_arguments"+number+");\n");

		return joinCode.toString();
	}
}
